import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
from .ash import apply_ash
from .spheric import SphericLinear
from .hyptorch import HyperbolicMLR, ToPoincare


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(BasicBlock, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(Bottleneck, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion * planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, in_channel=3, zero_init_residual=False):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(in_channel, 64, kernel_size=3, stride=1, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        #self.pm = ToPoincare(
        #    c=args.c_ball, 
        #    train_x=args.train_origin, 
        #    train_c=args.train_c, 
        #    ball_dim=args.feat_dim,
        #    riemannian=True
        #)
        #self.sph_layer = SphericLinear(512, 512, bias=False,)
				  #delta=delta, scaling=scaling, radius=radius, lrable=lrable,
				  #range_type=range_type, angle_type=angle_type, sign_type=sign_type)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves
        # like an identity. This improves the model by 0.2~0.3% according to:
        # https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for i in range(num_blocks):
            stride = strides[i]
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x, layer=100):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avgpool(out)
        if 'ash_method' in self.__dir__():
            out = apply_ash(out, method=getattr(self, 'ash_method'))
        #now out dim: batch_size, 512, 1, 1
        out = torch.flatten(out, 1) #start_dim = 1
        #out = self.sph_layer(out)
        return out

    # function to extact a specific feature
    def intermediate_forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)              
        return out

    # function to extact the multiple features
    def feature_list(self, x):
        out_list = []
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out_list.append(out)
        return out_list


def resnet18(**kwargs):
    return ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet34(**kwargs):
    return ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)


def resnet50(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(**kwargs):
    return ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)


model_dict = {
    'resnet18': [resnet18, 512],
    'resnet34': [resnet34, 512],
    'resnet50': [resnet50, 2048],
    'resnet101': [resnet101, 2048],
}


class LinearBatchNorm(nn.Module):
    """Implements BatchNorm1d by BatchNorm2d, for SyncBN purpose"""
    def __init__(self, dim, affine=True):
        super(LinearBatchNorm, self).__init__()
        self.dim = dim
        self.bn = nn.BatchNorm2d(dim, affine=affine)

    def forward(self, x):
        x = x.view(-1, self.dim, 1, 1)
        x = self.bn(x)
        x = x.view(-1, self.dim)
        return x

class SupCEResNet(nn.Module):
    """encoder + classifier"""
    def __init__(self, name='resnet18', normalize = False,  num_classes=10):
        super(SupCEResNet, self).__init__()
        model_fun, dim_in = model_dict[name]
        self.encoder = model_fun()
        self.fc = nn.Linear(dim_in, num_classes)
        self.normalize = normalize

    def forward(self, x):
        features = self.encoder(x)
        if self.normalize: 
            features =  F.normalize(features, dim=1)
        return self.fc(features)

class SupCEHeadResNet(nn.Module):
    """encoder + head"""
    def __init__(self, args, multiplier=1):
        super(SupCEHeadResNet, self).__init__()
        model_fun, dim_in = model_dict[args.model]
        if args.in_dataset == 'ImageNet-100':
            model = models.resnet34(pretrained=True)
            for name, p in model.named_parameters():
                if not name.startswith('layer4'):
                    p.requires_grad = False
            modules=list(model.children())[:-1] # remove last linear layer
            self.encoder =nn.Sequential(*modules)
        else:
            self.encoder = model_fun()
        self.fc = nn.Linear(args.feat_dim, args.n_cls) 
        #self.fc = SphericLinear(args.feat_dim, num_classes)
        self.multiplier = multiplier

        if args.head == 'linear':
            self.head = nn.Linear(dim_in, args.feat_dim)
        elif args.head == 'mlp':
            self.head = nn.Sequential(
                nn.Linear(dim_in, dim_in),
                nn.ReLU(inplace=True),
                #SphericLinear(dim_in, args.feat_dim)
                nn.Linear(dim_in, args.feat_dim)
            )
        elif args.head == 'hypb':
            self.head = nn.Sequential(
                # nn.Linear(dim_in, dim_in),
                # nn.ReLU(inplace=True),
                nn.Linear(dim_in, args.feat_dim),
                ToPoincare(
                    c=args.c_ball, 
                    train_x=args.train_origin, 
                    train_c=args.train_c, 
                    ball_dim=args.feat_dim,
                    riemannian=False
                )
            )
            #self.fc = HyperbolicMLR(ball_dim=args.feat_dim, n_classes=args.n_cls, c=args.c_ball)

        elif args.head == 'manifold':
            self.head = nn.Sequential(
                nn.Linear(dim_in, dim_in),
                nn.ReLU(inplace=True),
                nn.Linear(dim_in, args.feat_dim)
            )
            self.head2 = nn.Sequential(
                # nn.Linear(dim_in, dim_in),
                # nn.ReLU(inplace=True),
                nn.Linear(dim_in, args.feat_dim),
                nn.ReLU(inplace=True),
                ToPoincare(
                    c=args.c_ball, 
                    train_x=args.train_origin, 
                    train_c=args.train_c, 
                    ball_dim=args.feat_dim,
                    riemannian=False
                )
            )
            self.heads = [self.head, self.head2]
            n_head = 2
            self.fc = nn.Linear(args.feat_dim*n_head, args.n_cls)
        #self.fc = nn.Linear(dim_in, args.n_cls) # dim_in

        if args.ash_method:
            setattr(self.encoder, 'ash_method', args.ash_method)
            print(f"Use {getattr(self.encoder, 'ash_method')}")
    
    
    def forward(self, x, multi=True):
        rep = self.head_forward(x, multi=multi)
        if multi:
            rep = torch.hstack(rep)
        return self.fc(rep) #self.fc(self.head_forward(x)


    def head_forward(self, x, normalize=False, multi=False):
        feat = self.encoder(x).squeeze()
        features = self.head(feat)
        features = F.normalize(features, dim=1)
        if multi:
            features = [features]
            for head in self.heads[1:]:
                features2 = head(feat)
                features2 = F.normalize(features2, dim=1)
                features.append(features2)
        return features
    
    def intermediate_forward(self, x, normalize=True):
        feat = self.encoder(x).squeeze()
        if normalize: 
            return F.normalize(feat, dim=1)
        else: 
            return feat
